(* $Id: texec.ml 885 2008-02-12 10:44:49Z isr $ *)

open Big_int
open Utils
open Ttype
open Vars
open Tlib


let isBigZero x : bool = (sign_big_int x) = 0
let ( +>> ) x y = add_big_int x y
let ( ->> ) x y = sub_big_int x y
let ( *>> ) x y = mult_big_int x y
let ( />> ) x y = div_big_int x y
let ( %>> ) x y = mod_big_int x y

let vvAdd vl vr   : varValue = match getVType vl, getVType vr with
    VD_FLOAT, _ | _, VD_FLOAT -> VFloat (gvf vl +. gvf vr)
  | VD_STR, VD_STR -> VStr (gvs vl ^ gvs vr)
  | _ -> VInt (gvi vl +>> gvi vr)
let vvMinus vl vr : varValue = 
  if iof vl vr then VFloat (gvf vl -. gvf vr) else VInt (gvi vl ->> gvi vr)
let vvTimes vl vr : varValue =
  if iof vl vr then VFloat (gvf vl *. gvf vr) else VInt (gvi vl *>> gvi vr)
let vvDiv vl vr   : varValue =
  if iof vl vr then VFloat (gvf vl /. gvf vr) else VInt (gvi vl />> gvi vr)
let vvMod vl vr   : varValue =
  VInt (gvi vl %>> gvi vr)
let vvUminus vr   : varValue =
  if is_vfloat vr then VFloat (-. gvf vr) else VInt (minus_big_int (gvi vr))
let vvEqual vl vr : bool = match getVType vl, getVType vr with
    VD_ARRAY, _ | _, VD_ARRAY -> false    (* TODO *)
  | VD_FUNC,  _ | _, VD_FUNC  -> false    (* TODO *)
  | VD_FLOAT, _ | _, VD_FLOAT -> (gvf vl =  gvf vr)
  | VD_STR, VD_STR  -> (gvs vl = gvs vr)
  | _ -> eq_big_int (gvi vl) (gvi vr)

let initArray v : varValue = VArray [v]
let addVv2V (VArray vl) vr : varValue = VArray (vl @ [vr])
let addV2Vv v (VArray vv)  : varValue = VArray (v :: vv)

let lenVar (v: varValue) : int = match v with
    (VArray a) -> List.length a
  | _ -> 0
let nthVar (v: varValue) (n: int) : varValue =
  match v with
    (VArray a) -> List.nth a n 
  | _ -> assert false


(**
 * Statement return value.
 *)
type retType = RET_NO | RET_RETURN | RET_BREAK | RET_NEXT
type stmtRet = {
    srType : retType;
    srValue  : varValue
}
let initStmtRet t v = { srType = t; srValue = v }
let initStmtNoRet = initStmtRet RET_NO vzero

(* Convert function parameters to real varDef list. *)
let rec getNewVarDefs (ss: string list) (vvs: varValue list) : varDef list =
  match ss, vvs with
    [], _      -> []
  | s::ss', [] -> 
      if s.[0] = '*' then [initVarDefWithName (cutHeadChar s) (getVArgs [])]
      else []
  | s::ss', vv::vvs' -> 
      if s.[0] = '*' then [initVarDefWithName (cutHeadChar s) (getVArgs vvs)]
      else getNewVarDefs ss' vvs' @ [initVarDefWithName s vv]


(* Help replaceVar function. *)
let rec replaceVar' v (obj: varValue) (ns: int list) (i: int) : varValue = 
  if i >= lenVar v then VArray [] 
  else match ns with
      n::[] -> if n = i
               then addV2Vv obj (replaceVar' v obj ns (i+1))
               else addV2Vv (nthVar v i) (replaceVar' v obj ns (i+1))
    | n::rs -> if n = i
               then addV2Vv (replaceVar' (nthVar v i) obj rs 0) (replaceVar' v obj ns (i+1))
               else addV2Vv (nthVar v i) (replaceVar' v obj ns (i+1))
    | _ -> assert false

(* Replace the element of e with obj. *)
let replaceVar e tb (obj: varValue) (ns: int list) : varValue =
    replaceVar' (findVDexp tb e).vdValue obj ns 0



(*
 * Evaluate the expression of element.
 * e.g. [1][2] returns ([1;2], newds)
 *)
let rec getArrayNums (vmng: varManager) (e: exp) : int list =
  match e.eType with
    N_ARRAY_R -> let ns = getArrayNums vmng (eGetLeft e) in
                 let (VInt i) = evaluateExp vmng (eGetRight e) in
                 ns @ [int_of_big_int i]
  | N_PRIMARY -> getArrayNums vmng (eGetRight e) 
  | _ -> []

(* Get a N_VAR node by an easy way. *)
and getVarExp e : exp = match e.eType with
    N_VAR -> e
  | N_PRIMARY -> getVarExp (eGetRight e)
  | _ -> getVarExp (eGetLeft e)

and getCmVarArray' (vmng: varManager) (e: exp) : retExpCalc =
  let v = evaluateExp vmng e in
  if N_CM_LIST = e.eType
  then v
  else initArray v

(*
 * initTypeStruct'' updates 'ty' contents.
 *)
and initTypeStruct'' vmng (ty: tyStruct) (dlist: stDecl list) : unit =
  match dlist with
    [] -> ()
  | d::rs -> begin
      match d with
        StData s -> begin
          let nty = initTypeStruct vmng s in
          ty.tyTySt <- (nty::ty.tyTySt);
          asgnVarDef vmng (initVarDefWithName nty.tyName (VTySt nty));
        end
      | StVar  v -> begin
          ty.tyVars <- (v::ty.tyVars)
        end
      | StDef  d -> begin
          ty.tyDefs <- (d::ty.tyDefs);
          asgnVarDef vmng (initDefFunc vmng d);
        end
      | StInit (s, e) -> begin
          let v = evaluateExp vmng e in
          ty.tyInits <- ((s, v)::ty.tyInits);
          asgnVarDef vmng (initVarDefWithName s v);
        end
      | StLabel _ -> () (* TODO? *) 
      end;
      initTypeStruct'' vmng ty rs

and initTypeStruct' a b c d e =
  {tyName = a;  tyInits = b; tyVars = c; tyDefs = d; tyTySt = e}

and initTypeStruct (vmng: varManager) (s: stData) : tyStruct =
  let ty = initTypeStruct' s.stName [] [] [] [] in
  let new_vmng = incVarsMng vmng in
  initTypeStruct'' new_vmng ty s.stDecls;
  decVarsMng new_vmng;
  ty

and callFunc estmt vs =
  let (st, args, r, vmng2) = getVarStmt @@ getExpValue estmt in
  let ads = getNewVarDefs (getStringsFromVarNodes args) vs in
  let newvmng = incVarsMng vmng2 in
  asgnVarDefList newvmng (ads @ r);
  let ret = execStmt newvmng st in
  ret.srValue

and procAssign vmng el v =
  let ev = getVarExp el in
  let es = getArrayNums vmng el in
  let vd = if List.length es = 0
             then initVarDefWithExp ev v
             else initVarDefWithExp ev (replaceVar ev vmng v es)
  in
  asgnVarDef vmng vd;
  v
  
(* exp handler. *)
and evaluateExp (vmng: varManager) (e: exp) : retExpCalc =
  match e.eType with
    N_ARRAY   -> if e.eRight = None
                 then VArray []
                 else getCmVarArray' vmng (eGetRight e)

  | N_ARRAY_R -> let lval = evaluateExp vmng  (eGetLeft e) in
                 let rval = evaluateExp vmng (eGetRight e) in
                 nthList (gvmi rval) (gva lval)

  | N_ASSIGN -> procAssign vmng (eGetLeft e) (evaluateExp vmng (eGetRight e))

  | N_CM_LIST -> let lval = evaluateExp vmng (eGetLeft e) in
                 let rval = evaluateExp vmng (eGetRight e) in
                 if N_CM_LIST = (eGetLeft e).eType
                 then addVv2V lval rval
                 else VArray [lval; rval]
    
  | N_DOT -> let lval = evaluateExp vmng (eGetLeft e) in
             let ev = getVarExp (eGetRight e) in
             getMember vmng lval (eGetLabel ev)

  | N_EQEQ   -> let lval = evaluateExp vmng (eGetLeft e)  in
                let rval = evaluateExp vmng (eGetRight e) in
                if vvEqual lval rval then vunit else vzero

  | N_PRIMARY -> evaluateExp vmng (eGetRight e)

  | N_CALL_FUNC -> begin
      let v' = if e.eRight = None
               then (dput "CALL with no argument"; VArray [])
               else evaluateExp vmng (eGetRight e) in
      let v  = if e.eRight <> None
               then if N_CM_LIST = (eGetRight e).eType
                    then gva v'
                    else [v']
                       else [] in
      if N_LAMBDA <> (eGetLeft e).eType && isLibFunc (eGetLeft e)
      then callLibFunc (eGetLeft e) v
      else callFunc (getExpValue @@ evaluateExp vmng (eGetLeft e)) v
    end

  | N_CURRY  -> begin
      let vs' = evaluateExp vmng (eGetRight e) in
      assert (N_CM_LIST = (eGetRight e).eType);
      let vs = gva vs' in
      let (v, args) = (List.hd vs, List.tl vs) in
      curryVStmt v args
    end

  | N_LAMBDA -> let st   = initStmt STMT_RETURN   None None e.eRight None in
                let cmpd = initStmt STMT_COMPOUND None None None     (Some [st]) in
                dput "initial lambda";
                VStmt (cmpd, e.eLeft, [], vmng)

  | N_IF     -> let rval = evaluateExp vmng (eGetRight e) in
                if not @@ vvEqual rval vzero 
                then evaluateExp vmng (eGetLeft e)
                else vzero

  | N_AND    -> if not @@ vvEqual (evaluateExp vmng (eGetLeft e)) vzero 
                then if not @@ vvEqual (evaluateExp vmng (eGetRight e)) vzero
                     then vunit else vzero
                else vzero

  | N_OR     -> if vvEqual (evaluateExp vmng (eGetLeft e)) vzero 
                then if vvEqual (evaluateExp vmng (eGetRight e)) vzero
                     then vzero else vunit
                else vunit

  | N_NOEQ   -> let lval = evaluateExp vmng (eGetLeft e) in
                let rval = evaluateExp vmng (eGetRight e) in
                if not @@ vvEqual lval rval then vunit else vzero

  | N_LTHAN  -> let lval = evaluateExp vmng (eGetLeft e) in
                let rval = evaluateExp vmng (eGetRight e) in
                if gvf lval < gvf rval then vunit else vzero

    (* Add a new variable to an array. *)
    (* TODO: code-sharing with N_ASSIGN.  *)
  | N_LTLT   -> begin
      let rval = evaluateExp vmng (eGetRight e) in
      let lval = evaluateExp vmng (eGetLeft e) in
      let nar = VArray (gva lval @ [rval]) in
      let ev = getVarExp (eGetLeft e) in
      let es = getArrayNums vmng (eGetLeft e) in
      let newv = if List.length es = 0
                 then initVarDefWithExp ev nar
                 else initVarDefWithExp ev (replaceVar ev vmng nar es)
      in
      asgnVarDef vmng newv;
      nar
    end

  | N_ADD    -> let lval = evaluateExp vmng (eGetLeft e) in
                let rval = evaluateExp vmng (eGetRight e) in
                vvAdd lval rval

  | N_ADD_ASSIGN ->
      let el = eGetLeft e in
      let lval = evaluateExp vmng el in
      let rval = evaluateExp vmng (eGetRight e) in
      procAssign vmng el (vvAdd lval rval)

  | N_MINUS  -> let lval = evaluateExp vmng (eGetLeft e) in
                let rval = evaluateExp vmng (eGetRight e) in
                vvMinus lval rval

  | N_TIMES  -> let lval = evaluateExp vmng (eGetLeft e) in
                let rval = evaluateExp vmng (eGetRight e) in
                vvTimes lval rval

  | N_DIV    -> let lval = evaluateExp vmng (eGetLeft e) in
                let rval = evaluateExp vmng (eGetRight e) in
                vvDiv lval rval

  | N_MOD    -> let lval = evaluateExp vmng (eGetLeft e) in
                let rval = evaluateExp vmng (eGetRight e) in
                vvMod lval rval

  | N_UMINUS -> let v = evaluateExp vmng (eGetRight e) in
                vvUminus v

  | N_DEFINED -> let ev = getVarExp (eGetRight e) in
                 let ok =  Hashtbl.mem tlib_prevar_table (eGetLabel ev) 
                        || existVDexp vmng ev in
                 if ok then vunit else vzero

    (* TODO: 'nameof' *)
  | N_NAMEOF -> let v = evaluateExp vmng (eGetRight e)
                in v

  | N_INT    -> dput ("N_INT: " ^ eGetLabel e);
                VInt (big_int_of_string (eGetLabel e))

  | N_STRING -> dput ("N_STRING: " ^ eGetLabel e);
                VStr (eGetLabel e)

  | N_FLOAT  -> dput ("N_FLOAT: " ^ eGetLabel e);
                VFloat (float_of_string (eGetLabel e))

  | N_VAR    -> begin
      let name = eGetLabel e in
      dput ("N_VAR: " ^ name);
      if name = cancel_name then VCncl
      else 
        try Hashtbl.find tlib_prevar_table name
        with Not_found -> (findVDexp vmng e).vdValue
    end

  | _ -> assert false

(* Execute 'each' statement. *)
and execEach vmng (eo: exp option) (vvs: varValue list) (st: stmt) : stmtRet =
  match vvs with
    [] -> dput "STMT_EACH: End"; initStmtNoRet
  | vv::vvs' -> begin
      assert (eo <> None);
      asgnVarDef vmng (initVarDefWithExp (fromJust eo) vv);
      let ret = execStmt vmng st in
      match ret.srType with
        RET_RETURN -> ret
      | RET_BREAK  -> initStmtNoRet
      | _          -> execEach vmng eo vvs' st
    end

(* Execute 'times' statement. *)
and execTimes vmng (cnt: int) (st: stmt) : stmtRet =
  assert (cnt >= 0);
  if cnt = 0 then begin dput "STMT_TIMES: End"; initStmtNoRet end
  else 
    let ret = execStmt vmng st in
    match ret.srType with
      RET_RETURN -> ret
    | RET_BREAK  -> initStmtNoRet
    | _          -> execTimes vmng (cnt-1) st

(* Execute and handle all statements. *)
and  execStmt vmng s : stmtRet =
  match s.sType with
    STMT_CALC  -> let v = evaluateExp vmng (sGetExp s) in
                  begin
                    dput "STMT_CALC:";
                    ignore(tlpt_puts [VArray [v]]);
                    initStmtRet RET_NO v
                  end

  | STMT_COMPOUND -> execManage (sGetComp s) vmng

  | STMT_DEF -> begin
      dput "STMT_DEF";
      asgnVarDef vmng (initDefFunc vmng s);
      initStmtNoRet
    end

  | STMT_EMPTY -> dput "STMT_EMPTY";
                  initStmtNoRet

  | STMT_EXP   -> let v = evaluateExp vmng (sGetExp s) in
                  dput "STMT_EXP:";
                  initStmtRet RET_NO v

  | STMT_IF    -> let v = evaluateExp vmng (sGetCond s) in
                  if not @@ isBigZero (gvi v)
                  then execStmt vmng (sGetNest s)
                  else begin
                    dput "STMT_IF: Failed";
                    initStmtNoRet
                  end
    
  | STMT_EACH  -> let v = evaluateExp vmng (sGetCond s) in
                  execEach vmng (s.sExp) (gva v) (sGetNest s)

  | STMT_VEACH -> let v = getCmVarArray' vmng (sGetCond s) in
                  execEach vmng (s.sExp) (gva v) (sGetNest s)

  | STMT_LOOP -> begin
      let ret = execStmt vmng (sGetNest s) in
      match ret.srType with
        RET_RETURN -> ret
      | RET_BREAK  -> initStmtNoRet
      | _          -> execStmt vmng s
    end

  | STMT_CHANGE_LANG -> begin
      let lang = eGetLabel @@ sGetCond s in
      let src = eGetLabel @@ sGetExp s in
      let f = !(Lang.findLang lang) in
      f vmng src;
      initStmtNoRet
    end

  | STMT_TIMES -> begin
      let v = evaluateExp vmng (sGetCond s) in
      let cnt = match getVType v with
          VD_INT   -> gvmi v
        | VD_ARRAY -> List.length (gva v)
        | _ -> assert false
      in
      execTimes vmng cnt (sGetNest s) 
    end

  | STMT_WHILE -> begin
      let v = evaluateExp vmng (sGetCond s) in
      if not @@ isBigZero (gvi v)
      then let ret = execStmt vmng (sGetNest s)  in
           match ret.srType with
             RET_RETURN -> ret
           | RET_BREAK  -> initStmtNoRet
           | _          -> execStmt vmng s
      else (dput "STMT_WHILE: Failed";
            initStmtNoRet)
    end

  | STMT_DOWHILE -> begin
      let ret = execStmt vmng (sGetNest s) in
      match ret.srType with
        RET_RETURN -> ret
      | RET_BREAK  -> initStmtNoRet
      | _ -> 
          let v = evaluateExp vmng (sGetCond s) in
          if not @@ isBigZero (gvi v)
             then execStmt vmng s
             else (dput "STMT_DOWHILE: Failed";
                   initStmtNoRet)
    end

  | STMT_BREAK -> dput "STMT_BREAK";
                  initStmtRet RET_BREAK vzero

  | STMT_NEXT -> dput "STMT_NEXT";
                 initStmtRet RET_NEXT vzero

  | STMT_RETURN ->
          if s.sExp == None
          then (dput "STMT_RETURN";
                initStmtRet RET_RETURN vzero)
          else let v = evaluateExp vmng (sGetExp s)
               in (dput "STMT_RETURN with val";
                   initStmtRet RET_RETURN v)

  | STMT_PRINT -> begin
      match (sGetNest s).sType with
        STMT_EMPTY  -> print_endline "";
                       initStmtNoRet
      | STMT_EXP    -> let e = sGetExp @@ sGetNest s in
                       let v = evaluateExp vmng e in
                       let ns = getStringFromTree e ^ " is " ^ printVal v in
                       print_endline ns;
                       initStmtNoRet
    end

    (* TODO: STMT_STRUCT *)
  | STMT_STRUCT -> begin
      let ty = initTypeStruct vmng (sGetStruct s) in
      let d = initVarDefWithName ty.tyName (VTySt ty) in
      asgnVarDef vmng d;
      initStmtNoRet
    end

  | _ -> assert false

(* Execute statements and check the return values. *)
and execManage (ss: stmt list) (vmng: varManager) : stmtRet =
  match ss with
    [] -> dput "STMT: END";
          initStmtNoRet
  | s::ss' -> let ret = execStmt vmng s
              in if ret.srType <> RET_NO
                 then ret
                 else execManage ss' vmng

(* The main function of this file. *)
let tlptexec (sts: stmt list) (vmng: varManager) : stmtRet = 
    execManage sts vmng
