(* $Id: tlib.ml 854 2008-02-10 14:54:02Z isr $ *)
open Big_int
open Terr
open Ttype

let getVArgs' (vs: varValue list) : varValue list =
  let (VArray args) = List.nth vs 0 in args

let rec getStr' (args: varValue list) : string = match args with
      [] -> ""
    | arg::[] -> showVal arg
    | arg::rs -> showVal arg ^ " " ^ getStr' rs

(*
 * tlpt predefined variables
 *)
let tlpt_FLOAT_MAX : varValue = VFloat max_float
let tlpt_FLOAT_MIN : varValue = VFloat min_float
let tlpt_PI : varValue = VFloat 3.14159265358979323846264338327950288


(*
 * tlpt library functions
 *)
let tlpt_exit (vs: varValue list) : varValue = match List.nth vs 0 with
    VInt   i -> raise (Exit (int_of_big_int i))
  | _ -> assert false

let tlpt_float (vs: varValue list) : varValue = match List.nth vs 0 with
    VInt   i -> VFloat (float_of_big_int i)
  | VFloat _ as v -> v
  | _ -> assert false

let tlpt_gets (vs: varValue list) : varValue = VStr (read_line ())

let tlpt_int (vs: varValue list) : varValue = match List.nth vs 0 with
    VInt _ as v -> v
  | VFloat f -> VInt (big_int_of_int (int_of_float f))
  | _ -> assert false

let tlpt_is_array (vs: varValue list) : varValue = match List.nth vs 0 with
    VArray _ -> vunit
  | _ -> vzero

let tlpt_is_float (vs: varValue list) : varValue = match List.nth vs 0 with
    VFloat _ -> vunit
  | _ -> vzero

let tlpt_is_func (vs: varValue list) : varValue = match List.nth vs 0 with
    VStmt _ -> vunit
  | _ -> vzero

let tlpt_is_int (vs: varValue list) : varValue = match List.nth vs 0 with
    VInt _ -> vunit
  | _ -> vzero

let tlpt_is_string (vs: varValue list) : varValue = match List.nth vs 0 with
    VStr _ -> vunit
  | _ -> vzero


let tlpt_p (vs: varValue list) : varValue =
  let vargs = getVArgs' vs in
  print_string (getStr' vargs);
  flush stdout;
  vzero

let tlpt_puts (vs: varValue list) : varValue =
  let vargs = getVArgs' vs in
  let rec puts args = match args with
     []      -> print_endline ""
   | arg::[] -> print_endline (showVal arg)
   | arg::rs -> print_endline (showVal arg); 
                puts rs
  in
  puts vargs;
  flush stdout;
  vzero

let tlpt_sleep (vs: varValue list) : varValue = match List.nth vs 0 with
    VInt i -> Unix.sleep (int_of_big_int i); vzero
  | _ -> assert false

let tlpt_string (vs: varValue list) : varValue =
  let v = List.nth vs 0 in
  match v with
    VStr _ -> v
  | _ -> VStr (showVal (List.nth vs 0))


(*
 * Handle the above things.
 *)
let tlib_prevar_table =
  Utils.create_hashtable 10 [
    ("FLOAT_MAX",   tlpt_FLOAT_MAX);
    ("FLOAT_MIN",   tlpt_FLOAT_MIN);
    ("PI",          tlpt_PI);
  ]


let tlib_func_table =
  Utils.create_hashtable 32 [
    ("__tlpt_exit",      tlpt_exit);
    ("__tlpt_float",     tlpt_float);
    ("__tlpt_gets",      tlpt_gets);
    ("__tlpt_int",       tlpt_int);
    ("__tlpt_is_array",  tlpt_is_array);
    ("__tlpt_is_float",  tlpt_is_float);
    ("__tlpt_is_func",   tlpt_is_func);
    ("__tlpt_is_int",    tlpt_is_int);
    ("__tlpt_is_string", tlpt_is_string);
    ("__tlpt_p",         tlpt_p);
    ("__tlpt_puts",      tlpt_puts);
    ("__tlpt_sleep",     tlpt_sleep);
    ("__tlpt_string",    tlpt_string);
  ]

let isLibFunc (e: exp) : bool =
  if e.eLabel = None then false
  else Hashtbl.mem tlib_func_table (eGetLabel e)

let callLibFunc (e: exp) (vs: varValue list) : varValue =
  try
    let f = Hashtbl.find tlib_func_table (eGetLabel e) in
    f vs
  with Not_found -> assert false

