(* $Id: tlpt.ml 875 2008-02-12 05:00:26Z isr $ *)

open Terr
open Ttype
open Vars
open Texec

let version = "1-10 (OCaml version)"
let appname = "tlpt"
let header = appname ^ " [" ^ version ^ "]"

let exepath = Sys.executable_name
let lib_dir = Filename.concat (Filename.dirname exepath) "lib"
let libraries : string list = [ 
  "pervasives.tl";  (* Because of core library, load it first. *)
  "array.tl";
  "num.tl";
  ]


(* 
 * Calc mode functions
 *)
(* If the file of argv[1] is non-existence, tlpt starts argument calc mode *)
let isArgCalcMode : bool =
  Array.length Sys.argv > 1 && Sys.file_exists Sys.argv.(1) = false

let isCalcStmt (sts: stmt list) : bool = 
  List.length sts = 1 && (List.nth sts 0).sType = STMT_CALC

let getArgString (args: string array) : string =
  (Array.fold_left (fun x y -> x ^ " " ^ y)  "" args)

let getArgCalcString (args: string array) : string =
  "puts(" ^ getArgString args ^ ");"


(* 
 * Parse, execute the script codes.
 *)
let parseCheck (lb: Lexing.lexbuf) : bool * stmt list =
  try 
    (true, Tparser.input Tlex.tulpeta lb)
  with Parsing.Parse_error ->
    (false, [])

let execScript (lb: Lexing.lexbuf) (vmng: varManager) : bool * int option =
  let rec exec defs =
    let pok, stmts = parseCheck lb in
    if pok
    then begin
      ignore(tlptexec stmts defs);
      if lb.Lexing.lex_eof_reached then (true, None)
      else exec vmng
    end
    else (false, None)
  in
  try
    exec vmng
  with Exit i -> (true, Some i)


(* 
 * Script mode functions
 *)
(* 
 * NOTE: Don't write directly "Sys.argv" here. 
 *       Otherwise, an exeption occurs when an user select calc mode.
 *)
let getScriptFile (args: string array) : in_channel =
  if Array.length args = 1
  then stdin
  else open_in args.(1)


(*
 * Common functions
 *)
let rec readLibraries' lpath vmng : unit =
  if Sys.file_exists (lpath) 
  then begin
    let inch = open_in lpath in 
    ignore(execScript (Lexing.from_channel inch) vmng);
    close_in inch
  end

let readLibraries : varManager =
  let vmng = Vars.initVarsMng in
  List.iter 
    (fun lname -> readLibraries' (Filename.concat lib_dir lname) vmng)
    libraries;
  vmng

let main () =
  let libdefs = readLibraries in
  if isArgCalcMode
  then begin
    (* 
     * For calling __tlpt_puts(), Calc mode permits only one expression.
     *)
    let chks = getArgString (Utils.cutHeadEle Sys.argv) in
    let (ok, sts) = parseCheck (Lexing.from_string chks) in
    if ok = false || isCalcStmt sts = false then begin
      prerr_endline "[Calc Mode]";
      prerr_endline "error: parser error occured";
      prerr_endline "the following is your code.";
      prerr_endline "";
      prerr_endline chks;
      prerr_endline "";
      prerr_endline "NOTE: Calc mode permits only one expression";
      exit 200
    end
    else 
      let s = getArgCalcString (Utils.cutHeadEle Sys.argv) in
      ignore(execScript (Lexing.from_string s) libdefs)
  end
  else begin
    print_endline header;
    let inch = getScriptFile Sys.argv in
    let pok, ex = execScript (Lexing.from_channel inch) libdefs in
    close_in inch;
    if pok 
    then if ex = None then () else exit (Utils.fromJust ex)
    else begin
      prerr_endline "error: parser error occured";
      exit 201
    end
  end

let _ = Printexc.print main ()
