(* $Id: ttype.ml 871 2008-02-12 04:16:53Z isr $ *)
open Big_int
open Utils

let cancel_name = "><"

(**
 * If one operator (e.g. '-', '()') have two meanings, type 'token' can't work anymore. 
 * So, define type 't_node'.
 *)
type t_node 
    = N_ADD
    | N_ADD_ASSIGN
    | N_AND
    | N_ARRAY
    | N_ARRAY_R     (* is N_ARRAY_REFERENCE *)
    | N_ASSIGN
    | N_CALL_FUNC
    | N_CANCEL
    | N_CM_DEF_ARG
    | N_CM_LIST
    | N_CURRY
    | N_DEFINED
    | N_DIV
    | N_DOT
    | N_EQEQ
    | N_FLOAT
    | N_IF
    | N_INT
    | N_LAMBDA
    | N_LTHAN
    | N_LTLT
    | N_MINUS
    | N_MOD
    | N_NAMEOF
    | N_NOEQ
    | N_OR
    | N_PRIMARY
    | N_STRING
    | N_TIMES
    | N_UMINUS
    | N_V_ARGS
    | N_VAR

type exp = {
    eType  : t_node;
    eLabel : string option;   (* var name or integer value *)
    eLeft  : exp option;    (* left node *)
    eRight : exp option     (* right node *)
}
let initExp a b c d = {eType = a; eLabel = b; eLeft = c; eRight = d }
let eGetLabel e = fromJust e.eLabel
let eGetLeft  e = fromJust e.eLeft
let eGetRight e = fromJust e.eRight

let rec getStringsFromVarNodes (eo: exp option) : string list =
  let gsfv' = getStringsFromVarNodes in
  match eo with
    None -> []
  | Some e -> match e.eType with
      N_VAR    -> [eGetLabel e]
    | N_V_ARGS -> ["*" ^ eGetLabel (eGetRight e)]
    | _ ->  gsfv' e.eLeft @ [] @ gsfv' e.eRight

let rec getArgNodeNum (e: exp) : int =
  match e.eType with
    N_CM_DEF_ARG -> 1 + getArgNodeNum (eGetLeft e)
  | _ ->  1
    
(*
 * This 'cs' has reversed values. e.g. [4,3,1]
 * 'cs' does NOT have '0' value.
 *)
let rec cutExpNodes' (cs: int list) (e: exp) (i: int) : exp option =
  if List.length cs = 0 then Some e
  else begin
    match e.eType with
      N_CM_DEF_ARG -> 
        if i = List.hd cs
        then cutExpNodes' (List.tl cs) (eGetLeft e) (i-1)
        else Some (initExp N_CM_DEF_ARG None
                     (cutExpNodes' cs (eGetLeft e) (i-1)) e.eRight)

    | _ -> if i = List.hd cs then None else Some e
  end

let cutExpNodes (eo: exp option) (cs: int list) : exp option =
  match eo with
    None -> None
  | Some e -> begin
      let len = getArgNodeNum e in
      let rcs = List.rev cs in
      assert (len >= List.length cs);
      if len = List.length cs then None else cutExpNodes' rcs e len
    end

type t_stmt
    = STMT_BREAK
    | STMT_CALC
    | STMT_CHANGE_LANG
    | STMT_COMPOUND
    | STMT_DEF
    | STMT_DOWHILE
    | STMT_EACH
    | STMT_EMPTY
    | STMT_EXP
    | STMT_IF
    | STMT_LOOP
    | STMT_NEXT
    | STMT_PRINT
    | STMT_RETURN
    | STMT_STRUCT
    | STMT_TIMES
    | STMT_VEACH
    | STMT_WHILE

type stmt = {
    sType : t_stmt;       (* statement type *)
    sCond : exp option;   (* condition. STMT_DEF(argument), STMT_IF and iteration statement use *)
    sNest : stmt option;  (* STMT_DEF, STMT_IF, STMT_WHILE, STMT_DOWHILE, STMT_EACH, STMT_PRINT use *)
    sExp  : exp option;   (* STMT_DEF(func name), STMT_EXP, STMT_PRINT, STMT_RETURN, STMT_EACH use *)
    sComp : stmt list option; (* STMT_COMPOUND use *)
    sStruct : stData option; (* STMT_STRUCT *)
}
(* defined struct type *)
and stData = {
  stName   : string;      (* struct name *)
  stDecls : stDecl list;
}
and stDecl =
    StData of stData
  | StVar of string
  | StDef of stmt
  | StLabel of string
  | StInit of string * exp

let initStmt' a b c d e f =
  {sType = a; sCond = b; sNest = c; sExp = d; sComp = e; sStruct = f}
let initStmt a b c d e = initStmt' a b c d e None
let initStmtStruct st = initStmt' STMT_STRUCT None None None None st
let sGetCond s = fromJust s.sCond
let sGetNest s = fromJust s.sNest
let sGetExp  s = fromJust s.sExp
let sGetComp s = fromJust s.sComp
let sGetStruct s = fromJust s.sStruct
let initStData a b = {stName = a; stDecls = b}

type labelValue = string * varValue
and tyStruct = {
  tyName   : string;
  mutable tyInits  : labelValue list;
  mutable tyVars   : string list;
  mutable tyDefs   : stmt list;
  mutable tyTySt   : tyStruct list;
}
(* instance object *)
and insObj = {
  inType : tyStruct;
  inVars : labelValue list;
}

and varTbl = (string, varDef) Hashtbl.t
and varManager = varTbl list
and vstmt = stmt * exp option * varDef list * varManager

and varValue =
    VArray of varValue list
  | VCncl   (* cancel *)
  | VFloat of float
  | VIns   of insObj (* instance *)
  | VInt   of big_int
    (* stmt is function, and exp is the argument list. *)
  | VStmt  of vstmt
  | VStr   of string
  | VTySt  of tyStruct

and varDef = {
    vdType  : eVarType;
    vdLabel : string;
    vdValue : varValue
}

and  eVarType =
    VD_ARRAY
  | VD_CNCL
  | VD_FLOAT
  | VD_FUNC
  | VD_INS
  | VD_INT
  | VD_STR
  | VD_TYST

let makeVarDefNames vds : string list =
  List.map (fun x -> x.vdLabel) vds

(* return the string of value *)
let rec showVal' f (vs: varValue list) : string =
  match vs with
    [] -> ""
  | v :: [] -> f v
  | v :: rs -> f v ^ ", " ^ showVal' f rs

and showVal (v: varValue) : string = match v with
    VArray vs   -> "[" ^ showVal' showVal vs ^ "]"
  | VFloat f    -> string_of_float f
  | VIns _      -> "INSTANCE OBJECT" (* TODO: Implement this? *)
  | VInt i      -> string_of_big_int i
  | VStmt _     -> "FUNCTION" (* TODO: Implement this? *)
  | VStr s      -> s
  | VTySt _     -> "TYPE OBJECT" (* TODO: Implement this? *)
  | _ -> assert false

and printVal (v: varValue) : string = match v with
    VArray vs   -> "[" ^ showVal' printVal vs ^ "]"
  | VFloat f    -> string_of_float f
  | VIns _      -> "INSTANCE OBJECT" (* TODO: Implement this? *)
  | VInt i      -> string_of_big_int i
  | VStmt _     -> "FUNCTION" (* TODO: Implement this? *)
  | VStr s      -> "\"" ^ s ^ "\""
  | VTySt _     -> "TYPE OBJECT" (* TODO: Implement this? *)
  | _ -> assert false

let getVarInt (v: varValue) : big_int = match v with
    VInt i -> i
  | _ -> assert false

(* NOTE: getVarFloat does float_of_int, too. *)
let getVarFloat (v: varValue) : float = match v with
    VFloat f -> f
  | VInt i -> float_of_big_int i
  | _ -> assert false

let getVarArray (v: varValue) : varValue list = match v with
    VArray a -> a
  | _ -> assert false

let getVarStmt (v: varValue) : vstmt =
  match v with
    VStmt (s, e, r, n) -> (s, e, r, n)
  | _ -> assert false

let getVarString (v: varValue) : string = match v with
    VStr s -> s
  | _ -> assert false

let gvmi x = int_of_big_int (getVarInt x)
let gvi = getVarInt
let gvf = getVarFloat
let gva = getVarArray
let gvs = getVarString

let is_vfloat v : bool = match v with
    VFloat _ -> true
  | _ -> false
let is_one_vfloat vl vr = is_vfloat vl || is_vfloat vr 
let iof = is_one_vfloat

let getVType (v: varValue) : eVarType = match v with
    VArray _ -> VD_ARRAY
  | VCncl    -> VD_CNCL
  | VFloat _ -> VD_FLOAT
  | VIns   _ -> VD_INS
  | VInt   _ -> VD_INT
  | VStmt  _ -> VD_FUNC
  | VStr   _ -> VD_STR
  | VTySt  _ -> VD_TYST

let vzero = VInt zero_big_int
let vunit = VInt unit_big_int


let initVarDef a b c = {vdType = a; vdLabel = b; vdValue = c}

(* type definition for the return value of 'evaluateExp'. *) 
type retExpCalc = varValue

let getExpValue (v: retExpCalc) : varValue = v 

let initVarDefWithExp e (v: varValue) : varDef =
  match v with
    VArray _ -> initVarDef VD_ARRAY (eGetLabel e) v
  | VCncl    -> assert false
  | VFloat _ -> initVarDef VD_FLOAT (eGetLabel e) v
  | VInt   _ -> initVarDef VD_INT   (eGetLabel e) v
  | VStmt  _ -> initVarDef VD_FUNC  (eGetLabel e) v
  | VStr   _ -> initVarDef VD_STR   (eGetLabel e) v
  | VTySt  _ -> assert false
  | VIns   i -> initVarDef VD_INS   (eGetLabel e) v

let initVarDefWithName (s: string) (v: varValue) : varDef =
  match v with
    VArray _ -> initVarDef VD_ARRAY s v
  | VCncl    -> initVarDef VD_CNCL  s v
  | VFloat _ -> initVarDef VD_FLOAT s v
  | VInt   _ -> initVarDef VD_INT   s v
  | VStmt  _ -> initVarDef VD_FUNC  s v
  | VStr   _ -> initVarDef VD_STR   s v
  | VTySt  _ -> initVarDef VD_TYST  s v
  | VIns   _ -> initVarDef VD_INS   s v

(* Get the varDef of function *)
let initDefFunc mng (s: stmt) : varDef =
  initVarDef VD_FUNC (eGetLabel (sGetExp s))
                     (VStmt ((sGetNest s), s.sCond, [], mng))

let rec getVArgs (vvs: varValue list) : varValue =
  match vvs with
    [] -> VArray []
  | vv::vvs' -> VArray (vv :: gva (getVArgs vvs'))
    

let rec getTypeMemberDef mng (n: string) (ss: stmt list) : varValue option =
  match ss with
    [] -> None
  | s::rs -> if eGetLabel (sGetExp s) = n 
             then Some (VStmt ((sGetNest s), s.sCond, [], mng))
             else getTypeMemberDef mng n rs

let rec getTypeMemberTySt (n: string) (ts: tyStruct list) : varValue option =
  match ts with
    [] -> None
  | t::rs -> if t.tyName = n then Some (VTySt t)
             else getTypeMemberTySt n rs
  
let rec getTypeMemberInits (n: string) (is: labelValue list) : varValue option =
  match is with
    [] -> None
  | (vn, v)::rs -> if vn = n then Some v
             else getTypeMemberInits n rs
  
let getTypeMemberManage mng (n: string) (t: tyStruct) : varValue =
  let o = getTypeMemberDef mng n t.tyDefs in
  if o <> None then fromJust o
  else
    let o = getTypeMemberTySt n t.tyTySt in
    if o <> None then fromJust o
    else
      let o = getTypeMemberInits n t.tyInits in
      if o <> None then fromJust o
      else vzero

let rec getInsMemberVar (n: string) (vs: labelValue list) : varValue option =
  match vs with
    [] -> None
  | (vn, v)::rs -> if vn = n then Some v
             else getInsMemberVar n rs

let addThisToDef (i: insObj) (vst: varValue) : varValue =
  let (st, eo, vds, n) = getVarStmt vst in
  let vd = initVarDefWithName "this" (VIns i) in
  VStmt (st, eo, vd::vds, n)

let getInsMemberManage mng (n: string) (i: insObj) : varValue =
  let o = getTypeMemberDef mng n i.inType.tyDefs in
  if o <> None then addThisToDef i (fromJust o)
  else
    let o = getInsMemberVar n i.inVars in
    if o <> None then fromJust o else vzero

let initInsObj a b = {inType = a; inVars = b}
let rec newInsObj' (ns: string list) : labelValue list =
  List.map (fun x -> (x, vzero)) ns

let newInsObj (v: varValue) : varValue =
  match v with
    VTySt t -> VIns (initInsObj t (newInsObj' t.tyVars))
  | _ -> assert false


let getMember mng (v: varValue) (n: string) : varValue =
  match v with
    VTySt st  -> if n = "new" then newInsObj v else getTypeMemberManage mng n st
  | VIns  ins -> getInsMemberManage mng n ins
  | _ -> assert false

let rec curryVStmt' nms args (larg: varDef option) : string list * varDef list =
  match nms, args with
    _, []     -> (nms, [])
  | [], a::rs -> 
      if larg = None then ([], [])
      else let v = initVarDefWithName (fromJust larg).vdLabel a in
           let (rnms, lvs) = curryVStmt' [] rs (Some v) in
           (rnms, v::lvs)

  | n::rn, a::ra -> begin
     let v = initVarDefWithName n a in
     match v.vdType with
       VD_CNCL -> 
         let (rnms, lvs) = curryVStmt' rn ra larg in
          (rnms, lvs)
     | _ when n.[0] = '*' ->
         let (rnms, lvs) = curryVStmt' nms ra (Some v) in
          (rnms, v::lvs)
     | _ ->
         let (rnms, lvs) = curryVStmt' rn ra (Some v) in
          (rnms, v::lvs)
    end
  | _ -> assert false

(*
 * 'i' starts from 1.
 *)
let rec cutNums i (ns: string list) (vns: string list) : int list =
  match ns with
    []    -> []
  | n::rs -> if List.mem n vns then i::cutNums (i+1) rs vns
             else cutNums (i+1) rs vns
    
let curryVStmt (v: varValue) (args: varValue list) : varValue =
  let (st, pars, bs, n) = getVarStmt v in
  let pnms = getStringsFromVarNodes pars in
  if pnms = [] then v
  else begin
    let (rnms, nbs) = curryVStmt' pnms args None in
    let cnums = cutNums 1 pnms (makeVarDefNames nbs) in
    VStmt (st, (cutExpNodes pars cnums), (nbs @ bs), n)
  end


let rec getStringFromTree (e: exp) : string = match e.eType with
    N_ADD        -> addLeftRight e " + "
  | N_ADD_ASSIGN -> addLeftRight e " += "
  | N_AND        -> addLeftRight e " and "
  | N_ARRAY      -> if e.eRight = None then "[]"
                    else "[" ^ (getStringFromTree @@ eGetRight e) ^ "]"
  | N_ARRAY_R    -> (addLeftRight e "[") ^ "]"
  | N_ASSIGN     -> addLeftRight e " = "
  | N_CALL_FUNC  -> if e.eRight = None
                    then (getStringFromTree @@ eGetLeft e) ^ "()"
                    else (addLeftRight e "(") ^ ")"
  | N_CM_LIST    -> addLeftRight e ", "
  | N_CM_DEF_ARG -> addLeftRight e ", "
  | N_CURRY      -> "curry(" ^ getStringFromTree (eGetRight e) ^ ")"
  | N_DEFINED    -> "defined " ^ getStringFromTree (eGetRight e)
  | N_DIV        -> addLeftRight e " / "
  | N_DOT        -> addLeftRight e "."
  | N_EQEQ       -> addLeftRight e " == "
  | N_FLOAT      -> eGetLabel e
  | N_IF         -> addLeftRight e " if "
  | N_INT        -> eGetLabel e
  | N_LAMBDA     -> begin
      match eGetLabel e with
        "{}"   -> "{ " ^ getLambda e ^ " }"
      | "(\\)" ->  "(\\ " ^ getLambda e ^ " )"
      | "lambda()" -> "lambda(" ^ getLambda e ^ ")"
      | _ -> assert false
    end
  | N_LTLT       -> addLeftRight e " << "
  | N_LTHAN      -> addLeftRight e " < "
  | N_MINUS      -> addLeftRight e " - "
  | N_MOD        -> addLeftRight e " % "
  | N_NAMEOF     -> "nameof " ^ getStringFromTree (eGetRight e)
  | N_NOEQ       -> addLeftRight e " != "
  | N_OR         -> addLeftRight e " or "
  | N_PRIMARY    -> "(" ^ getStringFromTree (eGetRight e) ^ ")"
  | N_STRING     -> "\"" ^ eGetLabel e ^ "\""
  | N_TIMES      -> addLeftRight e " * "
  | N_UMINUS     -> "-" ^ getStringFromTree (eGetRight e)
  | N_V_ARGS     -> "*" ^ getStringFromTree (eGetRight e)
  | N_VAR        -> eGetLabel e
  | _            -> assert false

and addLeftRight (e: exp) (mid: string) : string =
  (getStringFromTree @@ eGetLeft e) ^ mid ^ (getStringFromTree @@ eGetRight e)

and getLambda' (eo: exp option) : string = match eo with
    None -> ""
  | Some e -> getStringFromTree e ^ " -> "

and getLambda (e: exp) : string =
  getLambda' e.eLeft ^ (getStringFromTree @@ eGetRight e)

