(* $Id: input.ml 886 2008-02-12 10:52:13Z isr $ *)

open Vm
open Tokens

(* Input to the whitespace VM.
   For convenience, three input characters 
       A => space, B => tab, C => either of CR/LF

Numbers are binary (A=0, B=1, C=terminator)
Strings are sequences of binary characters, terminated by C.

We have:

* Stack instructions (Preceded by A)
     Push (Integer)    A
     Dup           CA
     Swap          CB
     Discard       CC

* Arithmetic (Preceded by BA)
     Plus          AA
     Minus         AB
     Times         AC
     Divide        BA
     Modulo        BB

* Heap access (Preceded by BB)
     Store         A
     Retrieve      B

* Control     (Preceded by C)
     Label String  AA
     Call Label    AB
     Jump Label    AC
     If Zero Label BA
     If Neg Label  BB
     Return        BC
     End           CC

* IO instructions (Preceded by BC)
     OutputChar    AA
     OutputNum     AB
     ReadChar      BA
     ReadNum       BB

*)

(* get file contents *)
let slurp_channel channel =
  let buffer_size = 4096 in
  let buffer = Buffer.create buffer_size in
  let string = String.create buffer_size in
  let chars_read = ref 1 in
  while !chars_read <> 0 do
    chars_read := input channel string 0 buffer_size;
    Buffer.add_string buffer (String.sub string 0 !chars_read)
  done;
  Buffer.contents buffer

let slurp_file filename =
  let channel = open_in_bin filename in
  let result =
    try slurp_channel channel
    with e -> close_in channel; raise e in
  close_in channel;
  result

let whole_file = slurp_file

let list_of_chars s = Array.to_list (Array.init (String.length s) (fun i -> s.[i]))


let rec tokenise = function
    [] -> []
  | x::xs when x = showT A -> A::(tokenise xs)
  | x::xs when x = showT B -> B::(tokenise xs)
  | x::xs when x = showT C -> C::(tokenise xs)
  | _::xs -> tokenise xs


let makeNumber t =
  let rec makeNumber' tk pow =
      match tk with
      [] -> 0
    | (A::rest) -> (makeNumber' rest (pow*2))
    | (B::rest) -> pow + (makeNumber' rest (pow*2))
  in
  let init t =
    try
      List.rev (List.tl (List.rev t))
    with _ -> []
  in
  if List.hd (List.rev t) = A
  then   makeNumber' (init t) 1
  else -(makeNumber' (init t) 1)

let rec makeString tks = match tks with
    [] -> ""
  | (t::ts) -> String.make 1 (showT t) ^ (makeString ts)
(*
    let fst = take 8 ts in
    let rest = drop 8 ts in
    (toEnum (makeNumber fst)):(makeString rest) *)


let parseNumber ts = 
  let rec parseNum' tks acc =
      match tks with
      (C::rest) -> (makeNumber acc, rest)  
    | (x::rest) -> parseNum' rest (x::acc)  
  in
  parseNum' ts []

let parseString ts = 
  let rec parseStr' tks acc =
      match tks with
      (C::rest) -> (makeString acc, rest)  
    | (x::rest) -> parseStr' rest (x::acc)  
  in
  parseStr' ts []


let rec parse tks = match tks with
    [] -> []
  | (A::A::xs) -> let (num, rest) = parseNumber xs in
                  (Push num)::(parse rest)
  | (A::C::A::xs) -> Dup::(parse xs)
  | (A::B::A::xs) -> let (num,rest) = parseNumber xs in
                    (Ref num)::(parse rest)
  | (A::B::C::xs) -> let (num,rest) = parseNumber xs in
                     (Slide num)::(parse rest)
  | (A::C::B::xs) -> Swap::(parse xs)
  | (A::C::C::xs) -> Discard::(parse xs)

  | (B::A::A::A::xs) -> (Infix Plus)::(parse xs)
  | (B::A::A::B::xs) -> (Infix Minus)::(parse xs)
  | (B::A::A::C::xs) -> (Infix Times)::(parse xs)
  | (B::A::B::A::xs) -> (Infix Divide)::(parse xs)
  | (B::A::B::B::xs) -> (Infix Modulo)::(parse xs)

  | (B::B::A::xs) -> Store::(parse xs)
  | (B::B::B::xs) -> Retrieve::(parse xs)

  | (C::A::A::xs) -> let (str,rest) = parseString xs in
                     (Label str)::(parse rest)
  | (C::A::B::xs) -> let (str,rest) = parseString xs in
                     (Call str)::(parse rest)
  | (C::A::C::xs) -> let (str,rest) = parseString xs in
                     (Jump str)::(parse rest)
  | (C::B::A::xs) -> let (str,rest) = parseString xs in
                     (If (Zero, str))::(parse rest)
  | (C::B::B::xs) -> let (str,rest) = parseString xs in
                     (If (Negative, str))::(parse rest)
  | (C::B::C::xs) -> Return::(parse xs)
  | (C::C::C::xs) -> End::(parse xs)

  | (B::C::A::A::xs) -> OutputChar::(parse xs)
  | (B::C::A::B::xs) -> OutputNum::(parse xs)
  | (B::C::B::A::xs) -> ReadChar::(parse xs)
  | (B::C::B::B::xs) -> ReadNum::(parse xs)

  | _ -> prerr_endline "Unrecognised input";
         exit 10

let execute (src: string) = 
  let tokens = tokenise (list_of_chars src) in
  let runtime = parse tokens in
  vm (initVM runtime [] [] [] 0)


let execute_file (fname: string) = 
  let src = whole_file fname in
  execute src
