(* $Id: vm.ml 885 2008-02-12 10:44:49Z isr $ *)

(* Stack machine for running whitespace programs *)
open List

type op   = Plus | Minus | Times | Divide | Modulo
type test = Zero | Negative

type label = string
type loc = int

type instruction =
       Push of int
     | Dup
     | Ref of int
     | Slide of int
     | Swap
     | Discard
     | Infix of op
     | Store
     | Retrieve
     | Label of label
     | Call  of label
     | Jump  of label
     | If of test * label
     | Return
     | OutputChar
     | OutputNum
     | ReadChar
     | ReadNum
     | End

type heap = int list
type program = instruction list
type stack = int list

(* use 'initVM', if you wanna create this object. *)
type vmState = {
  vmProg     : program;
  vmVStack   : stack; (* V is Value *)
  vmCStack   : stack; (* C is Call *)
  vmMemory   : heap;
  vmPCounter : int  (* ? *)
}
let initVM a b c d e =
  { vmProg = a; vmVStack = b; vmCStack = c; vmMemory = d; vmPCounter = e }


(* Heap management *)
let rec store x n heap = match x, n, heap with
    x, 0, (h::hs) -> x::hs
  | x, n, (h::hs) -> let nhp = store x (n-1) hs in
                     h::nhp
  | x, 0, [] -> x::[]
  | x, n, [] -> let nhp = store x (n-1) [] in
                0::nhp

(* Digging out labels from wherever they are *)

let rec findLabel' lb prg i = match lb, prg, i with
    (* TODO: show l *)
    l, [], _ -> print_endline ("Undefined label foo");
                exit 40
  | m, ((Label l)::xs), i -> if l = m then i else findLabel' m xs (i+1)
  | m, (_::xs), i ->  findLabel' m xs (i+1)

let findLabel l p = findLabel' l p 0

let rec vm v = 
  let instr = nth v.vmProg v.vmPCounter in
  (* List.map (fun x -> print_int x; print_string ", ") v.vmVStack; *)
  doInstr (initVM v.vmProg v.vmVStack v.vmCStack v.vmMemory (v.vmPCounter+1)) instr

(* Running individual instructions *)
and doInstr v ins =
  match (v.vmProg, v.vmVStack, v.vmCStack, v.vmMemory, v.vmPCounter), ins with
    (prog, st, cs, hp, pc), (Push n) ->
      vm (initVM prog (n::st) cs hp pc)
  | (prog, (n::st), cs, hp, pc), Dup ->
      vm (initVM prog (n::n::st) cs hp pc)
  | (prog, st, cs, hp, pc), (Ref i) ->
      vm (initVM prog ((nth st i)::st) cs hp pc)
  | (prog, (n::st), cs, hp, pc), (Slide i) ->
      let rec drop lst = function 
          0 -> lst
        | n -> drop (tl lst) (n-1)
      in
      vm (initVM prog (n::(drop st i)) cs hp pc)
  | (prog, (n::m::st), cs, hp, pc), Swap ->
      vm (initVM prog (m::n::st) cs hp pc)
  | (prog, (n::st), cs, hp, pc), Discard ->
      vm (initVM prog st cs hp pc)
  | (prog, (y::x::st), cs, hp, pc), (Infix op) ->
      let doOp o l r =
          match o with
          Plus   -> l + r
        | Minus  -> l - r
        | Times  -> l * r
        | Divide -> l / r
        | Modulo -> l mod r
      in
      vm (initVM prog ((doOp op x y)::st) cs hp pc)
  | (prog, (n::st), cs, hp, pc), OutputChar ->
      print_char (char_of_int n);
      flush stdout;
      vm (initVM prog st cs hp pc)
  | (prog, (loc::st), cs, hp, pc), ReadChar ->
      let ch  = input_char stdin in
      let nhp = store (int_of_char ch) loc hp in
      vm (initVM prog st cs nhp pc)
  | (prog, (loc::st), cs, hp, pc), ReadNum ->
      let str = input_line stdin in
      let num = int_of_string str in
      let nhp = store num loc hp in
      vm (initVM prog st cs nhp pc)
  | (prog, (n::st), cs, hp, pc), OutputNum ->
      print_int n;
      flush stdout;
      vm (initVM prog st cs hp pc)
  | (prog, st, cs, hp, pc), (Label _) ->
      vm (initVM prog st cs hp pc)
  | (prog, st, cs, hp, pc), (Call l) ->
      let loc = findLabel l prog in
      vm (initVM prog st (pc::cs) hp loc)
  | (prog, st, cs, hp, pc), (Jump l) ->
      let loc = findLabel l prog in
      vm (initVM prog st cs hp loc)
  | (prog, (n::st), cs, hp, pc), (If (t, l)) ->
      let test ty num =
          match ty with
          Zero     -> num = 0
        | Negative -> num < 0
      in
      if (test t n)
      then let loc = findLabel l prog in
           vm (initVM prog st cs hp loc)
      else vm (initVM prog st cs hp pc)
  | (prog, st, (c::cs), hp, pc), Return ->
      vm (initVM prog st cs hp c)
  | (prog, (n::loc::st), cs, hp, pc), Store ->
      let nhp = store n loc hp in
      vm (initVM prog st cs nhp pc)
  | (prog, (loc::st), cs, hp, pc), Retrieve ->
      let retrieve n heap =
          nth heap n
      in
      let v = retrieve loc hp in
      vm (initVM prog (v::st) cs hp pc)
  | (prog, st, cs, hp, pc), End ->
      ()
  (* TODO: print the value of 'i' *)
  | _, i -> print_endline ("Can't do foo");
            exit 50

